#include <stdio.h>
#include <unistd.h>
#include <stdlib.h>
#include <fcntl.h>
#include <errno.h>
#include <string.h>
#include <nprobe.h>
 
 
 
int cycles = 0;
int overruns = 0;
char * sample;
 
 
// Perform the work of consuming a single Data Recording sample from NightProbe.
//
int
work (FILE * ofile, np_handle h, np_header * hdr) {
   np_item * i;
   int status;
   int which;
  
   // Read one sample, which may contain data for multiple processes
   // and variables.
   //
   status = np_read (h, sample);
   if (status <= 0) {
      return status;
   }
  
   cycles++;
  
   fprintf (ofile, "Sample %d\n", cycles);
   for (i = hdr->items; i; i = i->link) {
      char buffer [1024];
      sprintf (buffer, "item: %s:", i->name);
      fprintf (ofile, "%-30s", buffer); // Nice formatting :-)
  
      // Display the value of each item.
      // For arrays, format each individual item.
      //
      for (which = 1; which <= i->count; ++which) {
         char * image = np_format (h, i, sample, which);
  
         if (image != NULL) {
            fprintf (ofile, " %s", image);
         } else {
            fprintf (ofile, "\n<error: %s>\n", np_error (h));
            return -1;
         }
  
         free (image);
      }
      fprintf (ofile, "\n");
   }
   fflush (ofile);
  
   return 1;
}
  
  
  
  
int
main (int argc, char *argv[])
{
   np_handle h;
   np_header hdr;
   np_process * p;
   np_item * i;
   int fd;
   int status;
   FILE * ofile = stdout;
  
  
   fd = 0; // stdin
  
   status = np_open (fd, &hdr, &h);
   if (status) {
      fprintf (stderr, "%s\n", np_error (h));
      exit(1);
   }
  
   sample = (char *) malloc(hdr.sample_size);
   if (sample == NULL) {
      fprintf (stderr, "insufficient memory to allocate sample buffer\n");
      exit(1);
   }
  
   for (p = hdr.processes; p; p = p->link) {
      if (p->pid >= 0) {
         fprintf (ofile, "process: %s (%d)\n", p->name, p->pid);
      } else {
         fprintf (ofile, "resource: %s (%s)\n", p->name, p->label);
      }
   }
   fprintf (ofile, "\n");
  
   for (i = hdr.items; i; i = i->link) {
      fprintf (ofile, "item: %s (%s), size=%d bits, count=%d, type=%d\n",
             i->name, i->process->name, i->bit_size, i->count, i->type);
   }
   fprintf (ofile, "\n");
  
   for (;;) {
      status = work (ofile, h, &hdr);
      if (status <= 0) break;
   }
  
   fprintf (ofile, "Data Recording done: %d cycles fired, %d overruns\n",
            cycles, overruns);
  
   if (ofile != stdout) {
      fclose (ofile);
   }
  
   if (status < 0) {
      fprintf (stderr, "%s\n", np_error (h));
   }
  
   np_close (h);
  
   // At this point, file descriptor 0 remains open, but is no
   // longer a NightProbe Data File/Stream.
}

