//
// this example will :
//  A)remove fbs 4745 if it's already there
//  B)create fbs 4745
//  C)connect fbs to real time clock 2c1
//  D)program RTC 2c1 to fire every 850ms
//  E)schedule ourselves on the FBS
//  F)set out soft overrun limit to 1
//  G)start performance monitoring and clear tables
//  H)start the FBS
//  I)loop in fbswait loop,
//    1)print cycle, fbswait status, soft and hard overrun counts
//      to the file "DEBUG"
//    2)force a soft overun on cycles 10 and 20
//    3)force a hard overun starting cycle 25 (for 4-5 cycles)
//  J)cleanup fbs, DEBUG file and exit
//
//  problems, let me know
//
//  scott.shaffer@ccur.com
//

#include <sys/types.h>
#include <sys/ipc.h>
#include <string.h>
#include <stdlib.h>

#if 1    // for RedHawk
#include <fbsched.h>
#else    // for PowerMax
#include <fbslib.h>
#endif

#include <stdio.h>
#include <time.h>

int mydifftime(struct timespec *start, struct timespec *finish)
{
	int sec,nsec;
        
	sec = (finish->tv_sec - start->tv_sec);
	nsec= (finish->tv_nsec - start->tv_nsec);
	if (nsec<0)
	{
		sec --;
		nsec += 1000000000;
	}
	return (sec * 1000000000 + nsec);
}

main()
{
	int	istat;
	int	cnt=0;
	FILE	*fp;
	int	my_fbs_id,my_fpid;
	struct	fbsconfig_ds fbs_buf;
	struct	fbssched_buf sched_buf;
	struct	soft_overrun_ds soft_buf;
	struct	soft_overrun_info_ds soft_info_buf;
	struct	pmqry_ds pm_buf;
	char	name[2048];
	struct	timespec start, finish;
	int	result[1000];

//	fp=fopen("fbs_test.log","w");

//
// See if an old FBS with our key exists and remove it
//
	fbs_buf.key    = 4745;
	fbs_buf.cycles =    0;
	if( fbsconfigure( &fbs_buf ) != -1 )   // there is an old one out there
	{
		fbsrunrtc( fbs_buf.fbs_id, 0 );              // stop the rtc 
		fbsdetach( fbs_buf.fbs_id );                 // detach interrupt
		if( fbsremove( fbs_buf.fbs_id, 1 ) == -1 )   // nuke it
			perror("fbsremove");
	}

//
// create an FBS with our key
//
	fbs_buf.key       = 4745 ;
	fbs_buf.cycles    =  10 ;
	fbs_buf.progs     =    4 ;
	fbs_buf.max       =    4 ;
	fbs_buf.reset     =   -1 ;
	fbs_buf.configflg = 0666|IPC_CREAT ;
	if( fbsconfigure( &fbs_buf ) == -1 )
	{  perror("fbsconfigure");  exit(-1); }
	my_fbs_id = fbs_buf.fbs_id;


//
// attach our FBS to a real-time clock on our RCIM
//
	if( fbsattach( my_fbs_id, "/dev/rrtc/2c1" ) == -1 )
	{  perror("fbsattach");   exit(-1); }
	if( fbssetrtc( my_fbs_id, 100, 1000 ) == -1 )
	{  perror("fbssetrtc");   exit(-1); }
	if( fbsrunrtc( my_fbs_id, 1 ) == -1 )
	{  perror("fbsrunrtc");   exit(-1); }


//
// put ourselves on the FBS
//
	sched_buf.version = FBSSCHED_BUF_V1 ;
	sched_buf.param   =               0 ;
	sched_buf.period  =               1 ;
	sched_buf.cycle   =               0 ;
	sched_buf.ab      =               0 ;
	if( fbsschedself( fbs_buf.fbs_id, "fbs", &sched_buf ) == -1 )
	{  perror("fbsschedself");   exit(-1); }
	my_fpid = sched_buf.fpid;

//
// added 1 soft overrun limit for us
//
	name[0]='\000';
	soft_buf.name_ptr   = &name[0];
	soft_buf.cpu        = -1;
	soft_buf.fpid       = my_fpid;
	soft_buf.soft_limit =  1;
	if( sched_pgm_set_soft_overrun_limit( fbs_buf.fbs_id, &soft_buf ) == -1 )
		perror("sched_pgm_set_soft_overrun");

//
//  enable the performance monitor and clear the tables
//
	if( pmmonitor( my_fbs_id, 1, -1 ) == -1 )
		perror("pmmonitor");
	if( pmclrtable( my_fbs_id, -1 ) == -1 )
		perror("pmclrtable");

//
//  start the FBS
//
	if( fbsintrpt( my_fbs_id, -1 ) == -1 )
	{  perror("fbsintrpt");   exit(-1); }

	memset(&result[0], 0, sizeof(int) * 1000);
	clock_gettime(CLOCK_REALTIME, &start);

	while( (istat=fbswait()) != -1 )
	{
		clock_gettime(CLOCK_REALTIME, &finish);
#if 0
//
//		check for soft overruns
//		(return status of 2 from fbswait indicates that fbswait
//		did not block this time and simply returned, AKA soft overrun)
//
		name[0]='\000';
		soft_info_buf.name_ptr = &name[0];
		soft_info_buf.cpu      = -1;
		soft_info_buf.fpid     = my_fpid;
		if( sched_pgm_soft_overrun_query(my_fbs_id, &soft_info_buf) == -1 )
			perror("pmqrypgm");

//
//  check for hard overruns
//  (no indication from fbswait() that a hard overrun happened)
//
		pm_buf.fpid = my_fpid;
		name[0]='\000';
		if( pmqrypgm(my_fbs_id, name, -1, &pm_buf) == -1 )
			perror("pmqrypgm");

		fprintf(stdout,"cycle %d : fbswait()==%d,  soft OR count=%d   hard OR count=%d\n",
		cnt,istat,soft_info_buf.soft_total,pm_buf.overruns);

		if((cnt == 10) || (cnt == 20))	// force a "soft" overrun
			sleep(1);

		if(cnt == 25)			// force a "hard" overrun
			sleep(5);

		if(cnt >= 1000)			// enough of a test
			break;
#endif

		result[cnt] = mydifftime(&start, &finish);
		cnt ++;
		start.tv_sec = finish.tv_sec;
		start.tv_nsec = finish.tv_nsec;
	}

	fbsrunrtc( my_fbs_id, 0 );              // stop the rtc
	fbsdetach( my_fbs_id );                 // detach interrupt
	if( fbsremove( my_fbs_id, 1 ) == -1 )   // nuke it
		perror("fbsremove");

	fp=fopen("fbs_test.log","w");
	if (fp != NULL)
	{
		for (cnt=0; cnt<1000; cnt++)
		{
			fprintf(fp, "%d, %d\r\n", cnt, result[cnt]);
			fflush(fp);
		}
		fclose(fp);
	}
	printf("fbs test is done!\n");
}
