/*
*/
#include <rtl.h>

static rtltsd_t RTL_TSD[_POSIX_THREAD_THREADS_MAX];
static rtl_fifo_t  RTL_FIFO[_POSIX_THREAD_THREADS_MAX];
static int RTL_NUMBER=0;
static int SNO_NUMBER=0;
static int Stop_flag=0;
static int global_setup_signal(int signo, void (*interrupt_handler)(int,siginfo_t *, void *))
{
        static struct sigaction newact;
        static sigset_t set,oset;

        if (sigprocmask(0,NULL,&set)==(-1))
        {
                return(-1);
        }
        sigdelset(&set,signo);
        if (sigprocmask(SIG_SETMASK,&set,&oset)==(-1))
        {
                return(-2);
        }

        sigemptyset(&newact.sa_mask);
        sigaddset(&newact.sa_mask,signo);
        newact.sa_sigaction = interrupt_handler;
        newact.sa_flags = SA_SIGINFO|SA_RESTART;
        if (sigaction(signo, &newact, 0)==(-1))
        {
                return(-3);
        }
        return(0);
}

static void sigint_handler(int signo, siginfo_t *siginfo, void *misc)
{
    if ((siginfo->si_pid==0)||(getpid()==siginfo->si_pid)|| (getppid()==siginfo->si_pid))
    {
        fprintf(stderr,"\nSIG%d DETECT\n",signo);
        Stop_flag = 1;
    }
}

void wait_module(void)
{
	while(Stop_flag==0) pause();
}

void init_valueue(void)
{
	int fifo,sno;
	sigset_t diset,oset;
	struct sched_param myparam;

	SNO_NUMBER=WAITSIGNAL+1;
	sigemptyset(&diset);
	sigaddset(&diset,WAITSIGNAL);
	for(sno=WAITSIGNAL+1;sno<=SIGRTMAX;sno++)
	{
		sigaddset(&diset,sno);
	}
        sigprocmask(SIG_BLOCK,&diset,&oset);
	memset(RTL_TSD,0,sizeof(RTL_TSD));
	memset(RTL_FIFO,0,sizeof(RTL_FIFO));
	RTL_NUMBER=0;
	for(fifo=0;fifo<_POSIX_THREAD_THREADS_MAX;fifo++)
	{
		RTL_FIFO[fifo].f_id = -1;
	}
        myparam.sched_priority = sched_get_priority_max(SCHED_FIFO);
        sched_setscheduler(getpid(),SCHED_FIFO,&myparam);
	mlockall(MCL_CURRENT|MCL_FUTURE);
	global_setup_signal(SIGINT,sigint_handler);
}


static int get_sno(void)
{
	int sno;
	sno = SNO_NUMBER;
	if (sno>=SIGRTMAX)	return(-1);
	++SNO_NUMBER;
	return(sno);
}

static int get_tno(pthread_t pid)
{
	int tno;

	if (RTL_NUMBER)
	{
		for (tno=0;tno<RTL_NUMBER;tno++)
		{
			if (pthread_equal(RTL_TSD[tno].pid,pid) )
			{
				return(tno);
			}
		}
	}
	tno = RTL_NUMBER;++RTL_NUMBER;
	RTL_TSD[tno].pid = pid;
	RTL_TSD[tno].tno = tno;
	return(tno);
}



int pthread_wait_np(void)
{
	sigset_t diset,eiset,oset;
	int ret,signo,tno;

	tno = get_tno(pthread_self());
	resume:
	{
		sigfillset(&diset);
		for(signo=WAITSIGNAL;signo<=SIGRTMAX;signo++)
		{
			sigdelset(&diset,signo);
		}
        	if (pthread_sigmask(SIG_BLOCK,&diset,&oset)==(-1))
        	{
			return (-1);
        	}

		sigemptyset(&eiset);
		sigaddset(&eiset,WAITSIGNAL);
		sigaddset(&eiset,RTL_TSD[tno].timer_sno);
		sigaddset(&eiset,SIGINT);
		sigaddset(&eiset,SIGIO);
		RTL_TSD[tno].state=RTL_SUSPEND;
//printf("%d:sigwait(sleep:%d)\n",tno,RTL_TSD[tno].timer_sno);
		ret = sigwait(&eiset,&signo);
//fprintf(stderr,"%d:sigwait(wakeup) signo %d\n",tno,signo);
		RTL_TSD[tno].state=RTL_RUN;
        	if (pthread_sigmask(SIG_SETMASK,&oset,&eiset)==(-1))
        	{
			return (-1);
        	}
	} 
#if 0
	if ((signo==WAITSIGNAL)&&(RTL_TSD[tno].request==RTL_SUSPEND) )
	{
		RTL_TSD[tno].request=0;
		//if (signo!=SIGINT) goto resume;
	}
#endif
//printf("pthread_wait_np(ret)\n");
	return(ret); 
}

hrtime_t clock_gethrtime(clockid_t clock)
{
        struct timespec nowtime;
        long long x;

        clock_gettime(clock,&nowtime);
        x = (long long)nowtime.tv_sec << 32;
        x |= (long long)nowtime.tv_nsec;
	return(x);
}

int rtl_cpu_exists (int cpu)
{
        cpuset_t* setp;
        char *cpustr;
        int ret,mask;

        setp = cpuset_alloc();
        cpuset_init(setp);
        ret = mpadvise (MPA_CPU_PRESENT,0,0,setp);
        if (ret!=MPA_FAILURE)
        {
                cpustr = cpuset_get_string(setp);
                mask=  strtol(cpustr, (char **)NULL, 16);
                free(cpustr);
		return(mask & cpu);
        }
        free(cpustr);
	return(-1);
}

int rtl_num_cpus(void)
{
        cpuset_t* setp;
        char *cpustr;
        int ret,cpu,mask;

        setp = cpuset_alloc();
        cpuset_init(setp);
        ret = mpadvise (MPA_CPU_ACTIVE,0,0,setp);
        if (ret!=MPA_FAILURE)
        {
                free(cpustr);
		return ( cpuset_count(setp) );
        }
        free(cpustr);
	return(-1);
}

hrtime_t gethrtime(void)
{
        struct timespec nowtime;
        long long x;

        clock_gettime(CLOCK_REALTIME,&nowtime);
        x = (long long)nowtime.tv_sec << 32;
        x |= (long long)nowtime.tv_nsec;
	return(x);
}

int pthread_make_periodic_np(pthread_t thread,hrtime_t start_time,hrtime_t period)
{
        static struct sigevent event;
        static timer_t timer;
        static struct itimerspec itspec;
	int tno,policy;
	void * val=NULL;
	struct sched_param param;

        pthread_getschedparam(pthread_self(),&policy,(struct sched_param *)&param);
        param.sched_priority = sched_get_priority_max(SCHED_FIFO);
        pthread_setschedparam(pthread_self(),SCHED_FIFO,(const struct sched_param *)&param);


	tno = get_tno(thread);
        /* Create the POSIX timer to generate signum */
        event.sigev_notify = SIGEV_SIGNAL;
	RTL_TSD[tno].timer_sno = event.sigev_signo = get_sno();
	if (event.sigev_signo < (WAITSIGNAL+1))
	{
                return (-1);
	}
        if (timer_create((clockid_t)CLOCK_REALTIME,&event,&timer)==(-1))
        {
		RTL_TSD[tno].timer_key=(timer_t)-1;
                return (-1);
        }
        /*      Set the initial delay and period of the timer.
        This also arms the timer */
        clock_gettime(CLOCK_REALTIME,&itspec.it_value);
	start_time+=10000LL;
        itspec.it_value.tv_sec = (unsigned long int)(start_time>>32);
        itspec.it_value.tv_nsec = (unsigned long int)(start_time & 0xFFFFFFFFLL);
        itspec.it_interval.tv_sec = (unsigned long int)(period>>32);
        itspec.it_interval.tv_nsec = (unsigned long int)(period & 0xFFFFFFFFLL);
        if (timer_settime( timer, TIMER_ABSTIME, &itspec, NULL)==(-1))
        {
		RTL_TSD[tno].timer_key=(timer_t)-1;
                return (-1);
        }
	RTL_TSD[tno].timer_key=timer;
	RTL_TSD[tno].state=RTL_RUN;
	sched_yield();
        return(0);
}

int pthread_wakeup_np(pthread_t thread)
{
	int tno;
	tno = get_tno(pthread_self());
	RTL_TSD[tno].state=RTL_RUN;
	return(pthread_kill(thread, WAITSIGNAL));
}

int pthread_suspend_np(pthread_t thread)
{
	int tno;
	tno = get_tno(pthread_self());
	if (pthread_equal(pthread_self(),thread) )
	{
		if (RTL_TSD[tno].timer_key!=((timer_t)-1))
			timer_delete(RTL_TSD[tno].timer_key);
	}
	else
	{
		return(pthread_kill(thread, WAITSIGNAL));
	}
}

static void notification(void *arg) /* スレッド開始関数 */
{
	unsigned int sno;
	unsigned long fifo;
	rtl_fifo_t  *tcb;
	int status=0,retval, fd_flags;

	fifo  = (unsigned long)arg;
	tcb = &RTL_FIFO[fifo];

	tcb->flag=1;
#if 0
	sigemptyset(&tcb->sigmask);
	sigaddset(&tcb->sigmask,WAITSIGNAL);
	for(sno=WAITSIGNAL+1;sno<=SIGRTMAX;sno++)
	{
		sigaddset(&tcb->sigmask,sno);
	}
        if (pthread_sigmask(SIG_BLOCK,&tcb->sigmask,NULL)==(-1))
        {
		return ;
        }
#endif
#if 0
        fd_flags = fcntl(tcb->f_id, F_GETFL);
        if(fd_flags == -1)
        {
                return;
        }
        if(fcntl(tcb->f_id, F_SETFL, fd_flags | O_NONBLOCK) == -1)
        {
                return;
        }
#endif
	while(tcb->flag)
	{
		tcb->timeout.tv_sec = 0;
		tcb->timeout.tv_usec = 500000;
		FD_ZERO(&tcb->readfds);
		FD_SET(tcb->f_id, &tcb->readfds);
		retval = select(FD_SETSIZE,&tcb->readfds, NULL, NULL, &tcb->timeout);
		if (retval>0)
		{
			if (FD_ISSET(tcb->f_id, &tcb->readfds))
			{
				tcb->handler(fifo);
			}
		}
	}
	pthread_exit((void*)&status);
}


long rtf_create_handler(unsigned long fifo,long (*handler)(unsigned long fifo) )
{
	rtl_fifo_t  *tcb;

	tcb = &RTL_FIFO[fifo];

        if (pthread_attr_init(&tcb->attributes))
        {
		return(-1);
        }
        if (pthread_attr_setscope(&tcb->attributes,PTHREAD_SCOPE_SYSTEM))
        {
		return(-1);
        }
        if (pthread_attr_setdetachstate(&tcb->attributes,PTHREAD_CREATE_DETACHED))
        {
		return(-1);
        }
        if (pthread_attr_setinheritsched(&tcb->attributes,PTHREAD_INHERIT_SCHED))
        {
		return(-1);
        }
	tcb->handler = handler;
	if (pthread_create(&tcb->pid,&tcb->attributes,(void*)notification,(void *)fifo) < 0)
	{
		return(-1);
	}
	return(0);
}

int rtf_create(unsigned int fifo,int size)
{
	rtl_fifo_t  *tcb;

	tcb = &RTL_FIFO[fifo];
	tcb->flag=0;
        sprintf(tcb->f_name,"/dev/rtf%d",fifo);
	if (size>4096)
	{
		return(-1);
	}
	if (mkfifo(tcb->f_name, 0666)<0) 
        {
            return(-1);
        }
	tcb->f_id = open(tcb->f_name,O_RDWR|O_NONBLOCK,0666);
        return(tcb->f_id);
	return(0);
}

int rtf_destroy(unsigned int fifo)
{
	rtl_fifo_t  *tcb;

	tcb = &RTL_FIFO[fifo];
        if (tcb->f_id!=(-1))
        {
                close(tcb->f_id);
		tcb->f_id = -1;
        }
        sprintf(tcb->f_name,"/dev/rtf%d",fifo);
	if (!access(tcb->f_name,F_OK)) 
	{
		unlink(tcb->f_name);
	}
	if (tcb->flag)
	{
		tcb->flag=0;
		if (tcb->pid) pthread_kill(tcb->pid,SIGINT);
		tcb->pid=0;
	}
	return(0);
}

int rtf_put(unsigned int fifo,char *buff,int count)
{
	rtl_fifo_t  *tcb;
	int ret;

	tcb = &RTL_FIFO[fifo];
  	ret = write(tcb->f_id,buff,count);
  	return(ret);
}

int rtf_get(unsigned int fifo,char *buff,int count)
{
	rtl_fifo_t  *tcb;

	tcb = &RTL_FIFO[fifo];
  	return(read(tcb->f_id,buff,count));
}
int rtf_flush(unsigned int fifo)
{
	rtl_fifo_t  *tcb;

	tcb = &RTL_FIFO[fifo];
  	return(tcflush(tcb->f_id,TCIOFLUSH));
}


int pthread_attr_setcpu_np(const pthread_attr_t *attr,int cpu)
{
        cpuset_t        *cpuset;
	int ret;

        cpuset = cpuset_alloc();
        cpuset_init(cpuset);
        cpuset_set_cpu(cpuset,cpu,1);/* CPU set */

	ret = pthread_attr_setaffinity_np(attr, sizeof(cpuset), cpuset);
        cpuset_free(cpuset);
	return(ret);
}

int pthread_attr_getcpu_np(const pthread_attr_t *attr,int *cpu)
{
        cpuset_t        *cpuset;
	char *str;
	int i,ret;


        cpuset = cpuset_alloc();
        cpuset_init(cpuset);

	ret = pthread_attr_getaffinity_np(attr, sizeof(cpuset), cpuset);
	*cpu=-1;
	for(i=cpuset_min_cpu();i<cpuset_max_cpu();i++)
	{
		 if (cpuset_get_cpu(cpuset,i)) *cpu = i;
	}
        cpuset_free(cpuset);
	return(ret);
}

int rtl_pthread_create(pthread_t *thread, pthread_attr_t *attr, void *(*start_routine)(void *), void *arg)
{
	struct sched_param param;
	pthread_attr_setschedpolicy(attr,SCHED_FIFO);
        param.sched_priority = sched_get_priority_max(SCHED_FIFO);
	pthread_attr_setschedparam(attr,&param);
	return(pthread_create(thread, attr, start_routine, arg));
}

int rtl_gpos_mknod(const char *pathname, int mode, int major, int minor)
{
        return(mknod(pathname,(mode_t)(mode|S_IFCHR),(dev_t)((major<<16)|(minor))));
}

