/*
	POSIX RedHawk BenchMark test program version 3.0
*/
#include <stdio.h>
#include <time.h>
#include <signal.h>
#include <errno.h>
#include <sched.h>
#include <sys/types.h>
#include <unistd.h>
#include <mpadvise.h>
#include <cpuset.h>
#include <sys/mman.h>
#include <stdlib.h>
#include <math.h>
#include <sys/rescntl.h>

#define MAXJITTER (30)
#define BOOTMAXJITTER (100)
#define USEC 100
#define MAXCOUNT 100001
#define TSC
extern int errno;
static BOOT_TEST=0;
static struct resched_var rv;

#ifdef TSC
unsigned long long int rdtsc() {
unsigned int lo, hi;
/* We cannot use "=A", since this would use %rax on x86_64 */
__asm__ __volatile__ ("rdtsc" : "=a" (lo), "=d" (hi));
return (unsigned long long int)hi << 32 | lo;
}

double mhz;
double get_clock(void)
{
        FILE *f = fopen("/proc/cpuinfo", "r");
        if (f == 0){
                perror("can't open /proc/cpuinfo\n");
                exit(1);
        }

        for ( ; ; ){
                int ret;
                char buf[1000];

                if (fgets(buf, sizeof(buf), f) == NULL){
                        fprintf(stderr, "FATAL: cannot locate cpu MHz in " "/proc/cpuinfo\n");
                        exit(1);
                }
                ret = sscanf(buf, "cpu MHz         : %lf", &mhz);
                if (ret == 1){
                        fclose(f);
                        return mhz;
                }
        }
}
#endif

void
readtime(struct timespec *nowtime)
{
#ifdef TSC
	unsigned long long T;
	T = rdtsc();
	nowtime->tv_nsec = T & 0xFFFFFFFF;
	nowtime->tv_sec = (T>>32) & 0xFFFFFFFF;
#else
	clock_gettime(CLOCK_REALTIME,nowtime);
#endif
}

void
adjust(struct timespec *start,struct timespec *finish, double *realtime)
{
#ifdef TSC
	unsigned long long t0,t1;
	t0 = (unsigned long long)start->tv_sec;
	t0 <<=32;
	t0 |= 0xFFFFFFFF & (unsigned long long)(start->tv_nsec);
	t1 = (unsigned long long)finish->tv_sec;
	t1 <<=32;
	t1 |= 0xFFFFFFFF & (unsigned long long)(finish->tv_nsec);
	*realtime = (double)(t1-t0)/mhz;
#else 
        int     sec,nsec;

        sec = finish->tv_sec - start->tv_sec;
        nsec = finish->tv_nsec - start->tv_nsec;
        if (nsec<0)
        {
                sec --;
                nsec += 1000000000;
        }
        *realtime = (double)(nsec/1000)+(double)(sec*1000000);
#endif
}

void timer_test(double jitter)
{
        int signum,status;
        static struct sigaction newact;
        static sigset_t set,oset;
	static siginfo_t info;
        static struct timespec T,t0[MAXCOUNT],S;
	static double real[MAXCOUNT],min,max,avr;
	static int count = 0;
	double r;

	get_clock();

printf("********************************************* \n");
printf("****  posix timer test wait for %d sec **** \n",(MAXCOUNT-1)*USEC/1000000);
printf("********************************************* \n");
        if (sigprocmask(0,NULL,&set)==(-1))
        {
                fprintf(stderr, "Cannot get sigprocmask - %s\n",strerror(errno));
                exit(1);
        }
        sigaddset(&set,SIGRTMIN);
        if (sigprocmask(SIG_SETMASK,&set,&oset)==(-1))
        {
                fprintf(stderr, "%s: Cannot set sigprocmask - %s\n", strerror(errno));
                exit(1);
        }
	count =0;
	resched_lock (&rv);
        readtime(&T);
        do
        {
		do
		{
        		readtime(&t0[count]);
                	adjust(&T,&t0[count],&r);
		} while(r<(USEC));
        	T=t0[count];
		count++;
        } while (count<MAXCOUNT);
	resched_unlock (&rv);
	min=999999999;
	avr =max=0;
        for(count=1;count<MAXCOUNT;count++)
	{
                adjust(&t0[count-1],&t0[count],&real[count]);
		avr += real[count];
		if (min> real[count]) min =  real[count];
		if (max< real[count]) max =  real[count];
	}
	avr /= (double)(MAXCOUNT-1);
	printf(" min %f avr %f max %f\t",min,avr,max);
	if (fabs(max-min)<=jitter)
	{
		printf("GOOD(fabs(%f)<%f micro sec)\n",(max-min),jitter);
	}
	else
	{
		printf("FAIL(fabs(%f)>%f micro sec)\n",(max-min),jitter);
	}
}

main(int argc, char **argv)
{
	static struct sched_param myparam;
	cpuset_t* setp;
	int ret;
        extern int errno;
        extern int optind;
        extern char *optarg;
        register int c;
        char    *progname;

        progname = argv[0];
        while((c = getopt(argc, argv, "B")) != EOF){
                switch(c) {
                        case 'B':
                                BOOT_TEST=1;
                                break;
                        default:
                                fprintf(stderr,
                                "usage: %s [-B] B:BOOT_CPU_TEST\n",
                                progname);
                                exit(1);
                }
        }
#if 0
        if (optind != argc) {
                for (c = optind; c < argc; c++) {
                        if (access(argv[c], 4)) {
                                fprintf(stderr,
                                "%s: \"%s\" %s\n",
                                 progname,argv[c],strerror(errno));
                                exit(1);
                        }
                }
        }
#endif
	resched_cntl (RESCHED_SET_VARIABLE, (char *)&rv);
	mlockall(MCL_CURRENT|MCL_FUTURE);

	setp = cpuset_alloc();
        myparam.sched_priority = sched_get_priority_max(SCHED_FIFO);
        sched_setscheduler(getpid(),SCHED_FIFO,&myparam);
	system("/usr/bin/shield -r");
#if 0
	system("/usr/bin/shield -c");
	system("/usr/bin/cpu");
	timer_test(MAXJITTER);
	printf("\n\n");
#endif
	system("/usr/bin/shield -p 1 -i 1");
	system("/usr/bin/shield -c");
	system("/usr/bin/cpu");
	cpuset_set_string(setp, "2");
	ret = mpadvise (MPA_PRC_SETBIAS,MPA_PID,0,setp);
	if (ret!=MPA_FAILURE)
        {
		printf("running 2nd CPU\n");
        }
	else
	{
		printf("Can not running 2nd CPU\n");
	}
	timer_test(MAXJITTER);

	if (BOOT_TEST)
	{
		cpuset_set_string(setp, "2");
		ret = mpadvise (MPA_PRC_SETBIAS,MPA_PID,0,setp);
		if (ret!=MPA_FAILURE)
        	{
			printf("running boot CPU\n");
        	}
		else
		{
			printf("Can not running boot CPU\n");
		}
		timer_test(BOOTMAXJITTER);
	}
	system("/usr/bin/shield -a 1");
	system("/usr/bin/shield -c");
	system("/usr/bin/cpu");
	cpuset_set_string(setp, "4");
	ret = mpadvise (MPA_PRC_SETBIAS,MPA_PID,0,setp);
	if (ret!=MPA_FAILURE)
        {
		printf("running 2nd CPU\n");
        }
	else
	{
		printf("Can not running 2nd CPU\n");
	}
	timer_test(MAXJITTER);

}
